/* Utilities for shohei and shoheiftpd */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdarg.h>
#include <string.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <dirent.h>

#include "shoheiutil.h"

static FILE *sock_in, *sock_out;

static struct {
	char *key, *xl;
} *dict = NULL;

static long nw = 0;

static void default_shohei_error(char *p)
{
	fprintf(stderr, "%s\n", p);
   	exit(EXIT_FAILURE);
}

void (*shohei_error)(char *) = default_shohei_error;

void *cmalloc(size_t n)
{
	void *p = malloc(n);
	if (p == NULL) (*shohei_error)(_("Out of memory"));
	return p;
}

char *dupstring(char *p)
{
	char *q = cmalloc(strlen(p)+1);
	strcpy(q, p);
	return q;
}

char *catstrings3(char *p1, char *p2, char *p3)
{
	char *q;

	if (p1 == NULL) p1 = "";
	if (p2 == NULL) p2 = "";
	if (p3 == NULL) p3 = "";
	q = cmalloc(strlen(p1)+strlen(p2)+strlen(p3)+1);
	sprintf(q, "%s%s%s", p1, p2, p3);
	return q;
}

void chomp(char *p)
{
	while (*p) {
		if (*p == '\r' || *p == '\n') {
			*p = '\0';
			return;
		}
		p++;
	}
}

int cstrcasecmp(const char *p, const char *q)
{
        int c;
 
        while (!(c = toupper(*p)-toupper(*q)) && *p) {
                p++;
                q++;
        }
        return c;
}

int cstrncasecmp(const char *p, const char *q, size_t n)
{
        size_t i = 0;
        int c = 0;

        while ((i < n) && !(c = toupper(*p)-toupper(*q)) && *p) {
                p++;
                q++;
                i++;
        }
        return c;
}

int csnprintf(char *str, size_t n, const char *format, ...)
{
	int m;
	va_list ap;
	va_start(ap, format);
#ifdef HAVE_VSNPRINTF
	m = vsnprintf(str, n, format, ap);
#else
	m = vsprintf(str, format, ap);
#endif
	va_end(ap);
	return m;
}

/* If fn.gz exists, uncompress it. Then continue like fopen.  */
FILE *zopen(char *fn, char *mode)
{
	char b[1000];
	struct stat statbuf;

	csnprintf(b, sizeof b, "%s.gz", fn);
	if (!stat(b, &statbuf)) {
		csnprintf(b, sizeof b, "gunzip %s.gz", fn);
		system(b);
	}
	return fopen(fn, mode);
}

static void load_dictionary(char *fn)
{
	FILE *fp;
	char b[1000], *p, *q;

	/* convert everything after language code to upper case */
	p = strchr(fn, '_');
	if (p) {
		q = p;
		while (*q) {
			*q = toupper(*q);
			q++;
		}
	}

	if (p) {
		q = strchr(p, '.');
	} else {
		q = NULL;
	}

	fp = zopen(fn, "r");	
	if (!fp && q) {
		/* try abridged file name */
		*q = '\0';
		fp = zopen(fn, "r");
	}
	if (!fp && p) {
		/* abridge some more */
		*p = '\0';
		fp = zopen(fn, "r");
	}
	if (!fp) return;

	while (fgets(b, sizeof b, fp)) {
		chomp(b);
		p = strchr(b, '\t');
		if (!p) continue;
		*p++ = 0;
		p += strspn(p, "\t");
		dict = realloc(dict, (nw+1)*(sizeof dict[0]));
		dict[nw].key = dupstring(b);
		dict[nw++].xl = dupstring(p);
	}
}

/* Translate a message string, return key if not found */
char *translate(char *key)
{
	long lower = 0, upper = nw-1;
	long i;
	int d;

	while (lower <= upper) {
		i = (lower+upper)/2;
		d = strcmp(key, dict[i].key);
		if (d == 0) return dict[i].xl;
		else if (d > 0) lower = i+1;
		else upper = i-1;
	}
	return key;
}

void i18n(char *lang)
{
	char p[1000];

	if (lang) {
		csnprintf(p, sizeof p, "%s/%s.dict", I18NDIR, lang);
		load_dictionary(p);
	}
}

static char *months[] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec", NULL
};

int mon_a2i(char *m)
{
	int i;
	for (i = 0; months[i]; i++)
		if (!strcmp(m, months[i])) break;
	return i+1;
}

char *mon_i2a(int m)
{
	return months[m-1];
}

char *short_date(char *ld)
{
	char sd[100], mon[10];
	struct tm t;

	while (*ld && !isdigit(*ld)) ld++;
	if (*ld == '\0') return dupstring("");

	sscanf(ld, " %d %s %d %d:%d:%d",
	       &t.tm_mday, mon, &t.tm_year,
	       &t.tm_hour, &t.tm_min, &t.tm_sec);
	t.tm_mon = mon_a2i(mon);
	csnprintf(sd, sizeof sd, "%d %s %.02d:%.02d\n",
	       t.tm_mday, mon_i2a(t.tm_mon),
	       t.tm_hour, t.tm_min);
	return dupstring(sd);
}

/*
Given a domain name, try to find the corresponding pop server.
First look for the name in SPOOLDIR/domain/@SERVER. If the
file does not exist but SPOOLDIR/domain does, return the
domain name. Otherwise return NULL.
*/
char *server_lookup(char *d)
{
	char b[1000], *p;
	FILE *fp;
	DIR *dp;

	csnprintf(b, sizeof b, "%s/%s", SPOOLDIR, d);
	dp = opendir(b);
	if (dp == NULL) return NULL;
	closedir(dp);
	strcat(b, "/@SERVER");
	fp = zopen(b, "r");
	if (fp && fgets(b, sizeof b, fp)) p = dupstring(b);
	else p = dupstring(d);
	chomp(p);
	if (fp) fclose(fp);
	return p;
}

void downcase(char *p)
{
	while (*p) {
		*p = tolower(*p);
		p++;
	}
}

/* write 1 line of data to sock_out */
void write_data(char *b)
{
	fputs(b, sock_out);
	fflush(sock_out);
}

/* read 1 line of data from socket s */
void read_data(char *b)
{
	fgets(b, 1000, sock_in);
	chomp(b);
}

void open_socket(char *host, int portnum)
{
	struct sockaddr_in sa;
	struct hostent *hp;
	int s;

	hp = gethostbyname(host);
	if (hp == NULL) hp = gethostbyname(host);
	if (hp == NULL) {
		fprintf(stderr, "Host = '%s', h_error = %d\n", host, h_errno);
		(*shohei_error)(_("gethostbyname returns NULL"));
	}
	memset(&sa, '\0', sizeof sa);
	bcopy(hp->h_addr, (char *)&sa.sin_addr, hp->h_length);
	sa.sin_family = hp->h_addrtype;
	sa.sin_port = htons(portnum);
	s = socket(hp->h_addrtype, SOCK_STREAM, 0);
	if (s < 0)
		(*shohei_error)(_("socket() failed"));
	if (connect(s, (struct sockaddr *)&sa, sizeof sa) < 0)
		(*shohei_error)(_("connect() failed"));
	sock_out = fdopen(s, "w");
	sock_in = fdopen(dup(s), "r");
	if (sock_out == NULL || sock_in == NULL)
		(*shohei_error)(_("Can't fdopen socket"));
}

void close_socket(void)
{
	if (sock_in) {
		fclose(sock_in);
		sock_in = NULL;
	}
	if (sock_out) {
		fclose(sock_out);
		sock_out = NULL;
	}
}

char *mimename(char *b, size_t bs, char *f, char *mimedefault)
{
	char m[1000], *e, *p;
	FILE *fp;

	p = strrchr(f, '/');
	if (p) f = p;
	p = strrchr(f, '.');
	if (p == NULL) p = f;
	else p++;
	csnprintf(m, sizeof m, "%s/mime.types", SHOHEIDIR);
	fp = fopen(m, "r");
	if (fp == NULL) {
		csnprintf(b, bs, "%s", mimedefault);
		return b;
	}

	while (fgets(b, bs, fp)) {
		if (b[0] == '#') continue;	/* comment */
		chomp(b);
		e = strtok(b, "\t ");
		if (e == NULL) continue;	/* empty line */
		/* b is now a mime name candidate */
		e = strtok(NULL, "\t ");
		if (e == NULL) continue;	/* no whitespace */
		/* e is now an extension list */
		while (e) {
			if (!cstrcasecmp(e, p)) {
				fclose(fp);
				return b;
			}
			e = strtok(NULL, "\t ");
		}
	}
	fclose(fp);
	csnprintf(b, bs, "%s", mimedefault);
	return b;
}

